import matplotlib.pyplot as plt
import numpy as np

from termcolor import colored

from .utils import _process_rnn_args, _save_rnn_fig
from .inspect_gen import _detect_nans


def rnn_histogram(model, _id, layer=None, input_data=None, labels=None,
                  mode='weights', equate_axes=1, data=None, configs=None,
                  **kwargs):
    """Plots histogram grid of RNN weights/gradients by kernel, gate (if gated),
       and direction (if bidirectional). Also detects NaNs and shows on plots.

    Arguments:
        model: keras.Model/tf.keras.Model.
        _id: str/int. int -> idx; str -> name
            idx: int. Index of layer to fetch, via model.layers[idx].
            name: str. Name of layer (full or substring) to be fetched.
                       Returns earliest match if multiple found.
        layer: keras.Layer/tf.keras.Layer. Layer whose gradients to return.
               Overrides `idx` and `name`
        input_data: np.ndarray & supported formats(1). Data w.r.t. which loss is
               to be computed for the gradient. Only for mode=='grads'.
        labels: np.ndarray & supported formats. Labels w.r.t. which loss is
               to be computed for the gradient. Only for mode=='grads'
        mode: str. One of: 'weights', 'grads'. If former, plots layer weights -
              else, plots layer weights grads w.r.t. `input_data` & `labels`.
        equate_axes: int: 0, 1, 2. 0 --> auto-managed axes. 1 --> kernel &
                     recurrent subplots' x- & y-axes lims set to common value.
                     2 --> 1, but lims shared for forward & backward plots.
                     Bias plot lims never affected.
        data: np.ndarray. Pre-fetched data to plot directly - e.g., returned by
              `get_rnn_weights`. Overrides `input_data`, `labels` and `mode`.
              `model` and layer args are still needed to fetch RNN-specific info.
        configs: dict. kwargs to customize various plot schemes:
            'plot':      passed to ax.imshow();    ax  = subplots axis
            'subplot':   passed to plt.subplots()
            'tight':     passed to fig.subplots_adjust(); fig = subplots figure
            'title':     passed to fig.suptitle()
            'annot':     passed to ax.annotate()
            'annot-nan': passed to ax.annotate() for `nan_txt`
            'save':      passed to fig.savefig() if `savepath` is not None.
    (1): tf.data.Dataset, generators, .tfrecords, & other supported TensorFlow
         input data formats

    kwargs:
        w: float. Scale width  of resulting plot by a factor.
        h: float. Scale height of resulting plot by a factor.
        show_borders:  bool.  If True, shows boxes around plots.
        show_xy_ticks: int/bool iter. Slot 0 -> x, Slot 1 -> y.
              Ex: [1, 1] -> show both x- and y-ticks (and their labels).
                  [0, 0] -> hide both.
        bins: int. Pyplot `hist` kwarg: number of histogram bins per subplot.
        savepath: str/None. Path to save resulting figure to. Also see `configs`.
               If None, doesn't save.

    Returns:
        (subplots_figs, subplots_axes) of generated subplots. If layer is
            bidirectional, len(subplots_figs) == 2, and latter's is also doubled.
    """

    w, h          = kwargs.get('w', 1), kwargs.get('h', 1)
    show_borders  = kwargs.get('show_borders', False)
    show_xy_ticks = kwargs.get('show_xy_ticks', [1, 1])
    show_bias     = kwargs.get('show_bias', True)
    bins          = kwargs.get('bins', 150)
    savepath      = kwargs.get('savepath', None)

    def _process_configs(configs, w, h, equate_axes):
        defaults = {
            'plot':    dict(),
            'subplot': dict(sharex=True, sharey=True, dpi=76, figsize=(9, 9)),
            'tight':   dict(),
            'title':   dict(weight='bold', fontsize=13, y=1.05),
            'annot':     dict(fontsize=12, weight='bold',
                              xy=(.90, .93), xycoords='axes fraction'),
            'annot-nan': dict(fontsize=12, weight='bold', color='red',
                              xy=(.05, .63), xycoords='axes fraction'),
            'save': dict(),
            }
        configs = configs or {}
        # override defaults, but keep those not in `configs`
        for key in defaults:
            defaults[key].update(configs.get(key, {}))
        kw = defaults.copy()

        if not equate_axes:
            kw['subplot'].update({'sharex': False, 'sharey': False})
        size = kw['subplot']['figsize']
        kw['subplot']['figsize'] = (size[0] * w, size[1] * h)
        return kw

    def _catch_unknown_kwargs(kwargs):
        allowed_kwargs = ('w', 'h', 'show_borders', 'show_xy_ticks', 'bins',
                          'savepath')
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise Exception("unknown kwarg `%s`" % kwarg)

    def _detect_and_zero_nans(matrix_data):
        nan_txt = _detect_nans(matrix_data)
        if nan_txt is not None:  # NaNs detected
            matrix_data[np.isnan(matrix_data)] = 0  # set NaNs to zero
        return matrix_data, nan_txt

    def _plot_bias(data, axes, direction_idx, bins, d, kw):
        gs = axes[0, 0].get_gridspec()
        for ax in axes[-1, :]:
            ax.remove()
        axbig = fig.add_subplot(gs[-1, :])

        matrix_data = data[2 + direction_idx * 3].ravel()
        matrix_data, nan_txt = _detect_and_zero_nans(matrix_data)
        _pretty_hist(matrix_data, bins, ax=axbig)

        d['gate_names'].append('BIAS')
        _style_axis(axbig, gate_idx=-1, kernel_type=None, nan_txt=nan_txt,
                    show_borders=show_borders, d=d, kw=kw)
        for ax in axes[-2, :].flat:
            # display x labels on bottom row above bias plot as it'll differ
            # per bias row not sharing axes
            ax.tick_params(axis='both', which='both', labelbottom=True)

    def _pretty_hist(matrix_data, bins, ax):
        # hist w/ looping gradient coloring & nan detection
        N, bins, patches = ax.hist(matrix_data, bins=bins, density=True)

        if len(matrix_data) < 1000:
            return  # fewer bins look better monochrome

        bins_norm = bins / bins.max()
        n_loops = 8   # number of gradient loops
        alpha = 0.94  # graph opacity

        for bin_norm, patch in zip(bins_norm, patches):
            grad = np.sin(np.pi * n_loops * bin_norm) / 15 + .04
            color = (0.121569 + grad * 1.2, 0.466667 + grad, 0.705882 + grad,
                     alpha)  # [.121569, .466667, ...] == matplotlib default blue
            patch.set_facecolor(color)

    def _get_axes_extrema(axes):
        axes = np.array(axes)
        is_bidir = len(axes.shape) == 3 and axes.shape[0] != 1
        x_new, y_new = [], []

        for direction_idx in range(1 + is_bidir):
            axis = np.array(axes[direction_idx]).T
            for type_idx in range(2):  # 2 == len(kernel_types)
                x_new += [np.max(np.abs([ax.get_xlim() for ax in axis[type_idx]]))]
                y_new += [np.max(np.abs([ax.get_ylim() for ax in axis[type_idx]]))]
        return max(x_new), max(y_new)

    def _set_axes_limits(axes, x_new, y_new, d):
        axes = np.array(axes)
        is_bidir = len(axes.shape) == 3 and axes.shape[0] != 1

        for direction_idx in range(1 + is_bidir):
            axis = np.array(axes[direction_idx]).T
            for type_idx in range(2):
                for gate_idx in range(d['n_gates']):
                    axis[type_idx][gate_idx].set_xlim(-x_new, x_new)
                    axis[type_idx][gate_idx].set_ylim(0,      y_new)

    def _style_axis(ax, gate_idx, kernel_type, nan_txt, show_borders, d, kw):
        if nan_txt is not None:
            ax.annotate(nan_txt, **kw['annot-nan'])

        is_gated = d['rnn_type'] in gated_types
        if gate_idx == 0:
            title = kernel_type + ' GATES' * is_gated
            ax.set_title(title, weight='bold')
        if is_gated:
            ax.annotate(d['gate_names'][gate_idx], **kw['annot'])

        if not show_borders:
            ax.set_frame_on(False)
        if not show_xy_ticks[0]:
            ax.set_xticks([])
        if not show_xy_ticks[1]:
            ax.set_yticks([])

    def _get_plot_data(data, direction_idx, type_idx, gate_idx, d):
        matrix_idx = type_idx + direction_idx * (2 + d['uses_bias'])
        matrix_data = data[matrix_idx]

        if d['rnn_type'] in d['gated_types']:
            start = gate_idx * d['rnn_dim']
            end   = start + d['rnn_dim']
            matrix_data = matrix_data[:, start:end]
        return matrix_data.ravel()

    def _make_subplots(show_bias, direction_name, d, kw):
        if not (d['uses_bias'] and show_bias):
            fig, axes = plt.subplots(d['n_gates'], 2, **kw['subplot'])
            axes = np.atleast_2d(axes)
            return fig, axes

        n_rows = 2 * d['n_gates'] + 1
        fig, axes = plt.subplots(n_rows, 2, **kw['subplot'])

        # merge upper axes pairs to increase height ratio w.r.t. bias plot window
        gs = axes[0, 0].get_gridspec()
        for ax in axes[:(n_rows - 1)].flat:
            ax.remove()
        axbigs1, axbigs2 = [], []
        for row in range(n_rows // 2):
            start = 2 * row
            end = start + 2
            axbigs1.append(fig.add_subplot(gs[start:end, 0]))
            axbigs2.append(fig.add_subplot(gs[start:end, 1]))
        axes = np.vstack([np.array([axbigs1, axbigs2]).T, [*axes.flat[-2:]]])

        if direction_name != []:
            fig.suptitle(direction_name + ' LAYER', **kw['title'])
        return fig, axes

    kw = _process_configs(configs, w, h, equate_axes)
    _catch_unknown_kwargs(kwargs)
    data, rnn_info = _process_rnn_args(model, _id, layer, input_data, labels,
                                       mode, data)
    d = rnn_info
    gated_types  = ('LSTM', 'GRU', 'CuDNNLSTM', 'CuDNNGRU')
    kernel_types = ('KERNEL', 'RECURRENT')
    d.update({'gated_types': gated_types, 'kernel_types': kernel_types})

    subplots_axes = []
    subplots_figs = []
    for direction_idx, direction_name in enumerate(d['direction_names']):
        fig, axes = _make_subplots(show_bias, direction_name, d, kw)
        subplots_axes.append(axes)
        subplots_figs.append(fig)

        for type_idx, kernel_type in enumerate(kernel_types):
            for gate_idx in range(d['n_gates']):
                ax = axes[gate_idx][type_idx]
                matrix_data = _get_plot_data(data, direction_idx,
                                             type_idx, gate_idx, d)

                matrix_data, nan_txt = _detect_and_zero_nans(matrix_data)
                _pretty_hist(matrix_data, bins=bins, ax=ax)

                _style_axis(ax, gate_idx, kernel_type, nan_txt, show_borders,
                            d, kw)
        if d['uses_bias'] and show_bias:
            _plot_bias(data, axes, direction_idx, bins, d, kw)

        if kw['tight']:
            fig.subplots_adjust(**kw['tight'])
        else:
            fig.tight_layout()

    if equate_axes == 2:
        x_new, y_new = _get_axes_extrema(subplots_axes)
        _set_axes_limits(subplots_axes, x_new, y_new, d)

    plt.show()
    if savepath is not None:
        _save_rnn_fig(subplots_figs, savepath, kw['save'])
    return subplots_figs, subplots_axes


def rnn_heatmap(model, _id, layer=None, input_data=None, labels=None,
                mode='weights', cmap='bwr', norm='auto', data=None,
                configs=None, **kwargs):
    """Plots histogram grid of RNN weights/gradients by kernel, gate (if gated),
       and direction (if bidirectional). Also detects NaNs and prints in console.

    Arguments:
        model: keras.Model/tf.keras.Model.
        _id: str/int. int -> idx; str -> name
            idx: int. Index of layer to fetch, via model.layers[idx].
            name: str. Name of layer (full or substring) to be fetched.
                       Returns earliest match if multiple found.
        layer: keras.Layer/tf.keras.Layer. Layer whose gradients to return.
               Overrides `idx` and `name`
        input_data: np.ndarray & supported formats(1). Data w.r.t. which loss is
               to be computed for the gradient. Only for mode=='grads'.
        labels: np.ndarray & supported formats. Labels w.r.t. which loss is
               to be computed for the gradient. Only for mode=='grads'.
        mode: str. One of: 'weights', 'grads'. If former, plots layer weights -
              else, plots layer weights grads w.r.t. `input_data` & `labels`.
        cmap: str. Pyplot cmap (colormap) kwarg for the heatmap. If None,
              uses 'bone' cmap (greyscale), suited for >=0 value heatmaps.
        norm: float list/tuple; str ('auto'); None. Normalizes colors to range
              between norm==(vmin, vmax), according to `cmap`.
                  Ex: `cmap`='bwr'('blue white red') -> all values <=vmin and
                  >=vmax will be shown as most intense  blue and red, and those
                  exactly in-between are shown as white.
              If 'auto', will normalize across all non-bias plots (per kernels,
              gates, and directions), zero-centered; however, if `absolute_value`
              is also True, sets vmin==vmax==None instead.
              If None, Pyplot handles norm.
        data: np.ndarray. Pre-fetched data to plot directly - e.g., returned by
              `get_rnn_weights`. Overrides `input_data`, `labels` and `mode`.
              `model` and layer args are still needed to fetch RNN-specific info.
        configs: dict. kwargs to customize various plot schemes:
            'plot':      passed to ax.imshow(); ax = subplots axis
            'plot-bias': passed to ax.imshow() for bias
            'subplot':   passed to fig.subplots(); fig = subplots figure
            'tight':     passed to fig.subplots_adjust()
            'title':     passed to fig.suptitle()
            'subtitle':  passed to ax.set_title()  of each subplot
            'xlabel':    passed to ax.set_xlabel() of each non-bias subplot
            'ylabel':    passed to ax.set_ylabel() of each non-bias subplot
            'colorbar':  passed to fig.colorbar()
            'save':      passed to fig.savefig() if `savepath` is not None.
    (1): tf.data.Dataset, generators, .tfrecords, & other supported TensorFlow
         input data formats

    kwargs:
        w: float. Scale width  of resulting plot by a factor.
        h: float. Scale height of resulting plot by a factor.
        show_borders:  bool. If True, shows boxes around plots.
        show_colorbar: bool. If True, shows one colorbar next to plot(s).
        show_bias:     bool. If True, includes plot for bias (if layer uses bias)
        gate_sep_width: float. Pyplot kwarg for `linewidth` in marking gate
              separations in gated RNNs (LSTM, GRU) w/ vertical lines.
        normalize: bool. If True, scales all values to lie between 0 & 1. Works
              well with a greyscale `cmap` (e.g. None). Applied after
              `absolute_value`.
        absolute_value: bool. If True, takes absolute value of all data before
              plotting. Works well with a greyscale `cmap` (e.g. None). Applied
              before `normalize`.
        savepath: str/None. Path to save resulting figure to. Also see `configs`.
               If None, doesn't save.

    Returns:
        (subplots_figs, subplots_axes) of generated subplots. If layer is
            bidirectional, len(subplots_figs) == 2, and latter's is also doubled.
    """

    w, h           = kwargs.get('w', 1), kwargs.get('h', 1)
    show_borders   = kwargs.get('show_borders',  True)
    show_colorbar  = kwargs.get('show_colorbar', True)
    show_bias      = kwargs.get('show_bias', True)
    gate_sep_width = kwargs.get('gate_sep_width', 1)
    normalize      = kwargs.get('normalize', False)
    absolute_value = kwargs.get('absolute_value', False)
    savepath       = kwargs.get('savepath', None)

    def _process_configs(configs, w, h):
        defaults = {
            'plot':      dict(interpolation='nearest'),
            'plot-bias': dict(interpolation='nearest'),
            'subplot':   dict(dpi=76, figsize=(14, 8)),
            'tight':     dict(),
            'title':     dict(weight='bold', fontsize=14, y=.98),
            'subtitle':  dict(weight='bold', fontsize=14),
            'xlabel':    dict(fontsize=12, weight='bold'),
            'ylabel':    dict(fontsize=12, weight='bold'),
            'colorbar':  dict(fraction=.03),
            'save':      dict(),
            }
        configs = configs or {}
        # override defaults, but keep those not in `configs`
        for key in defaults:
            defaults[key].update(configs.get(key, {}))
        kw = defaults.copy()

        size = kw['subplot']['figsize']
        kw['subplot']['figsize'] = (size[0] * w, size[1] * h)
        return kw

    def _catch_unknown_kwargs(kwargs):
        allowed_kwargs = ('w', 'h', 'show_borders', 'show_colorbar',
                          'show_bias', 'gate_sep_width', 'normalize',
                          'absolute_value', 'savepath')
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise Exception("unknown kwarg `%s`" % kwarg)

    def _process_data(data, absolute_value, normalize):
        if absolute_value:
            data = np.abs(data)
        if normalize:
            for idx in range(len(data)):
                data[idx] -= np.min(data[idx])
                if np.max(data[idx]) == 0:
                    data[idx] = -data[idx]
                else:
                    data[idx] /= np.max(data[idx])
        return data

    def _get_style_info(cmap, norm, d):
        def _make_common_norm(data):
            idxs = [0, 1] + [2, 3] * (len(data) == 4) + [3, 4] * (len(data) == 6)
            return np.max([np.max(np.abs(data[idx])) for idx in idxs])

        if cmap is None:
            cmap = plt.cm.bone

        if norm == 'auto':
            if absolute_value:
                vmin, vmax = None, None
            else:
                vmax = _make_common_norm(data)
                vmin = -vmax
        elif norm is None:
            vmin, vmax = None, None
        else:
            vmin, vmax = norm

        if d['gate_names'][0] != '':
            gate_names_str = '(%s)' % ', '.join(d['gate_names'])
        else:
            gate_names_str = ''

        return cmap, vmin, vmax, gate_names_str

    def _print_nans(nan_txt, matrix_data, kernel_type, gate_names, rnn_dim):
        if gate_names[0] == '':
            print(kernel_type, end=":")
            nan_txt = nan_txt.replace('\n', '')
            print(colored(nan_txt, 'red'))
        else:
            print('\n' + kernel_type + ":")
            n_gates = len(gate_names)
            gates_data = []
            for gate_idx in range(n_gates):
                start, end = rnn_dim*gate_idx, rnn_dim*(gate_idx + 1)
                gates_data.append(matrix_data[..., start:end])

            for gate_name, gate_data in zip(gate_names, gates_data):
                nan_txt = _detect_nans(gate_data)
                if nan_txt is not None:
                    nan_txt = nan_txt.replace('\n', '')
                    print(gate_name + ':', colored(nan_txt, 'red'))

    def _detect_and_print_nans(matrix_data, kernel_type, d):
        nan_txt = _detect_nans(matrix_data)
        if nan_txt is not None:
            _print_nans(nan_txt, matrix_data, kernel_type,
                        d['gate_names'], d['rnn_dim'])

    def _style_axis(ax, type_idx, kernel_type, show_borders, is_vector, d, kw):
        if is_vector:
            ax.set_xticks([])

        if d['gate_sep_width'] != 0:
            lw = d['gate_sep_width']
            [ax.axvline(d['rnn_dim'] * gate_idx - .5, linewidth=lw, color='k')
             for gate_idx in range(1, d['n_gates'])]

        ax.set_title(kernel_type, **kw['subtitle'])
        ax.set_xlabel(d['gate_names_str'], **kw['xlabel'])
        y_label = ['Channel units', 'Hidden units'][type_idx]
        ax.set_ylabel(y_label, **kw['ylabel'])

        if not show_borders:
            ax.set_frame_on(False)

    def _make_subplots(show_bias, direction_name, d, kw):
        if not (d['uses_bias'] and show_bias):
            fig, axes = plt.subplots(1, 2, **kw['subplot'])
            axes = np.atleast_2d(axes)
            return fig, axes

        fig, axes = plt.subplots(10, 2, **kw['subplot'])

        # merge upper 11 axes to increase height ratio w.r.t. bias plot window
        gs = axes[0, 0].get_gridspec()
        for ax in axes.flat[:18]:
            ax.remove()
        axbig1 = fig.add_subplot(gs[:8, 0])
        axbig2 = fig.add_subplot(gs[:8, 1])
        axes = np.array([[axbig1, axbig2], [*axes.flat[-2:]]])
        axes = np.atleast_2d(axes)

        if direction_name != []:
            fig.suptitle(direction_name + ' LAYER', **kw['title'])
        return fig, axes

    _catch_unknown_kwargs(kwargs)
    kw = _process_configs(configs, w, h)
    data, rnn_info = _process_rnn_args(model, _id, layer, input_data, labels,
                                       mode, data, norm)
    d = rnn_info
    d['gate_sep_width'] = gate_sep_width

    data = _process_data(data, absolute_value, normalize)
    cmap, vmin, vmax, d['gate_names_str'] = _get_style_info(cmap, norm, d)
    kernel_types = ['KERNEL', 'RECURRENT']

    subplots_axes = []
    subplots_figs = []
    for direction_idx, direction_name in enumerate(d['direction_names']):
        fig, axes = _make_subplots(show_bias, direction_name, d, kw)
        subplots_axes.append(axes)
        subplots_figs.append(fig)

        ez = lambda *x: enumerate(zip(*x))
        for type_idx, (kernel_type, ax) in ez(kernel_types, axes.flat):
            w_idx = type_idx + direction_idx * (2 + d['uses_bias'])
            matrix_data = data[w_idx]
            _detect_and_print_nans(matrix_data, kernel_type, d)

            is_vector = matrix_data.ndim == 1
            if is_vector:
                matrix_data = np.expand_dims(matrix_data, -1)

            if 'aspect' not in kw['plot']:
                aspect = 20 / len(matrix_data) if is_vector else 'auto'
                kw['plot']['aspect'] = aspect
            img = ax.imshow(matrix_data, cmap=cmap, vmin=vmin, vmax=vmax,
                            **kw['plot'])
            _style_axis(ax, type_idx, kernel_type, show_borders, is_vector,
                        d, kw)

        if is_vector:
            plt.subplots_adjust(right=.7, wspace=-.4)
        if show_colorbar:
            fig.colorbar(img, ax=axes[0, :], **kw['colorbar'])

        if d['uses_bias'] and show_bias:
            for ax in axes[-1, :]:
                ax.remove()
            gs = axes[0, 0].get_gridspec()
            axbig = fig.add_subplot(gs[-1, :])

            data_idx = 2 + direction_idx * (2 + d['uses_bias'])
            weights_viz = np.atleast_2d(data[data_idx])

            if 'aspect' not in kw['plot-bias']:
                dim = weights_viz.shape[1]
                kw['plot-bias']['aspect'] = dim / 100 if dim > 100 else .5
            axbig.imshow(weights_viz, cmap=cmap, vmin=vmin, vmax=vmax,
                         **kw['plot-bias'])
            # Styling
            if d['gate_sep_width'] != 0:
                lw = d['gate_sep_width']
                [axbig.axvline(d['rnn_dim'] * gate_idx - .5, linewidth=lw,
                               color='k') for gate_idx in range(1, d['n_gates'])]
            axbig.set_frame_on(False)
            axbig.set_title('BIAS', **kw['subtitle'])
            axbig.get_yaxis().set_ticks([])

        if kw['tight']:
            fig.subplots_adjust(**kw['tight'])

    plt.show()
    if savepath is not None:
        _save_rnn_fig(subplots_figs, savepath, kw['save'])
    return subplots_figs, subplots_axes



import numpy as np
from pathlib import Path
from ._backend import WARN, NOTE


def _validate_args(_id, layer=None):
    def _ensure_list(_id, layer):
        # if None, leave as-is
        _ids, layer = [[x] if not isinstance(x, (list, type(None))) else x
                       for x in (_id, layer)]
        # ensure external lists unaffected
        _ids, layer = [x.copy() if isinstance(x, list) else x
                       for x in (_ids, layer)]
        return _ids, layer

    def _ids_to_names_and_idxs(_ids):
        names, idxs = [], []
        for _id in _ids:
            if not isinstance(_id, (str, int, tuple)):
                tp = type(_id).__name__
                raise ValueError("unsupported _id list element type: %s" % tp
                                 + "; supported are: str, int, tuple")
            if isinstance(_id, str):
                names.append(_id)
            else:
                if isinstance(_id, int):
                    idxs.append(_id)
                else:
                    assert all(isinstance(x, int) for x in _id)
                    idxs.append(_id)
        return names or None, idxs or None

    def _one_requested(_ids, layer):
        return len(layer or _ids) == 1  # give `layer` precedence

    if (_id is not None and layer is not None):
        print(WARN, "`layer` will override `_id`")

    _ids, layer = _ensure_list(_id, layer)
    if _ids is None:
        names, idxs = None, None
    else:
        names, idxs = _ids_to_names_and_idxs(_ids)
    return names, idxs, layer, _one_requested(_ids, layer)


def _process_rnn_args(model, _id, layer, input_data, labels, mode,
                      data=None, norm=None):
    """Helper method to validate `input_data` & `labels` dims, layer info args,
       `mode` arg, and fetch various pertinent RNN attributes.
    """

    from .inspect_gen import get_layer, get_gradients
    from .inspect_rnn import get_rnn_weights

    def _validate_args_(_id, layer, input_data, labels, mode, norm, data):
        _validate_args(_id, layer)

        if data is not None:
            got_inputs = (input_data is not None) or (labels is not None)
            if got_inputs:
                print(NOTE, "`data` will override `input_data`, `labels`, "
                      "and `mode`")
            if not isinstance(data, list):
                raise Exception("`data` must be a list of kernel & gate matrices")

            if not (isinstance(data[0], np.ndarray) or isinstance(data[0], list)):
                raise Exception("`data` list elements must be numpy arrays "
                                + "or lists")
            elif isinstance(data[0], list):
                if not isinstance(data[0][0], np.ndarray):
                    raise Exception("`data` list elements' elements must be "
                                    + "numpy arrays")

        if mode not in ['weights', 'grads']:
            raise Exception("`mode` must be one of: 'weights', 'grads'")
        if mode == 'grads' and (input_data is None or labels is None):
            raise Exception("must supply input_data and labels for mode=='grads'")
        if mode == 'weights' and (input_data is not None or labels is not None):
            print(NOTE, "`input_data` and `labels will` be ignored for "
                  "`mode`=='weights'")

        is_iter = (isinstance(norm, list) or isinstance(norm, tuple) or
                   isinstance(norm, np.ndarray))
        is_iter_len2 = is_iter and len(norm)==2
        if (norm is not None) and (norm != 'auto') and not is_iter_len2:
            raise Exception("`norm` must be None, 'auto' or iterable ( "
                            + "list, tuple, np.ndarray) of length 2")

    _validate_args_(_id, layer, input_data, labels, mode, norm, data)
    if layer is None:
        layer = get_layer(model, _id)
    rnn_type = _validate_rnn_type(layer, return_value=True)

    gate_names = _rnn_gate_names(rnn_type)
    n_gates  = len(gate_names)
    is_bidir = hasattr(layer, 'backward_layer')
    rnn_dim  = layer.layer.units if is_bidir else layer.units
    direction_names = ['FORWARD', 'BACKWARD'] if is_bidir else [[]]
    if 'CuDNN' in rnn_type:
        uses_bias = True
    else:
        uses_bias  = layer.layer.use_bias if is_bidir else layer.use_bias

    if data is None:
        if mode=='weights':
            data = get_rnn_weights(model, _id, as_tensors=False,
                                   concat_gates=True)
        else:
            data = get_gradients(model, None, input_data, labels,
                                 layer=layer, mode='weights')

    rnn_info = dict(rnn_type=rnn_type, gate_names=gate_names,
                    n_gates=n_gates, is_bidir=is_bidir,
                    rnn_dim=rnn_dim, uses_bias=uses_bias,
                    direction_names=direction_names)
    return data, rnn_info


def _validate_rnn_type(rnn_layer, return_value=False):
    if hasattr(rnn_layer, 'backward_layer'):
        rnn_type = type(rnn_layer.layer).__name__
    else:
        rnn_type = type(rnn_layer).__name__

    supported_rnns = ['LSTM', 'GRU', 'CuDNNLSTM', 'CuDNNGRU',
                      'SimpleRNN', 'IndRNN']
    if rnn_type not in supported_rnns:
        raise Exception("unsupported RNN type `%s` - must be one of: %s" % (
                        rnn_type, ', '.join(supported_rnns)))
    if return_value:
        return rnn_type


def _rnn_gate_names(rnn_type):
    return {'LSTM':      ['INPUT',  'FORGET', 'CELL', 'OUTPUT'],
            'GRU':       ['UPDATE', 'RESET',  'NEW'],
            'CuDNNLSTM': ['INPUT',  'FORGET', 'CELL', 'OUTPUT'],
            'CuDNNGRU':  ['UPDATE', 'RESET',  'NEW'],
            'SimpleRNN': [''],
            'IndRNN':    [''],
            }[rnn_type]


def _filter_duplicates_by_keys(keys, *data):
    def _second_index(ls, k):
        return [i for i, x in enumerate(ls) if x == k][1]

    collected = []
    for k in keys:
        if k in collected:
            for i in range(len(data)):
                data[i].pop(_second_index(keys, k))
            keys.pop(keys.index(k))
        collected.append(k)
    if isinstance(data, tuple) and len(data) == 1:
        data = data[0]
    return keys, data


def _save_rnn_fig(figs, savepath, kwargs):
    if len(figs) == 1:
        figs[0].savefig(savepath)
        return

    _dir = str(Path(savepath).parent)
    ext = Path(savepath).suffix
    basename = Path(savepath).stem
    names = [basename + '_0', basename + '_1']

    for fig, name in zip(figs, names):
        fig.savefig(Path(_dir).joinpath(name, ext), **kwargs)



import os
from termcolor import colored


TF_KERAS = os.environ.get("TF_KERAS", '0') == '1'
WARN = colored("WARNING:", 'red')
NOTE = colored("NOTE:", 'blue')


if TF_KERAS:
    import tensorflow.keras.backend as K
    print(NOTE, "`sample_weights` & `learning_phase` not yet supported "
          "for `TF_KERAS`, and will be ignored (%s.py)" % __name__)
else:
    import keras.backend as K



from .inspect_gen import get_layer
from .utils import _validate_args, _validate_rnn_type
from ._backend import K, TF_KERAS, WARN


def get_rnn_weights(model, _id, layer=None, as_tensors=False, concat_gates=True):
    """Retrievers RNN layer weights.

    Arguments:
        model: keras.Model/tf.keras.Model.
        idx: int. Index of layer to fetch, via model.layers[idx].
        name: str. Name of layer (can be substring) to be fetched. Returns
              earliest match if multiple found.
        layer: keras.Layer/tf.keras.Layer. Layer whose gradients to return.
               Overrides `idx` and `name`.
        as_tensors: If True, returns weight tensors instead of array values.
               NOTE: in Eager, both are returned.
        concat_gates: If True, returns kernel weights are signle concatenated
               matrices, instead of individual per-gate weight lists.
    """

    names, idxs, *_ = _validate_args(_id, layer)
    name = names[0] if names is not None else None
    idx  = idxs[0]  if idxs  is not None else None

    if layer is None:
        layer = get_layer(model, name or idx)
    rnn_type = _validate_rnn_type(layer, return_value=True)
    IS_CUDNN = 'CuDNN' in rnn_type

    if hasattr(layer, 'backward_layer'):
        l = layer
        forward_cell  = l.forward_layer  if IS_CUDNN else l.forward_layer.cell
        backward_cell = l.backward_layer if IS_CUDNN else l.backward_layer.cell

        forward_cell_weights  = _get_cell_weights(forward_cell,  as_tensors,
                                                  concat_gates)
        backward_cell_weights = _get_cell_weights(backward_cell, as_tensors,
                                                  concat_gates)
        return forward_cell_weights + backward_cell_weights
    else:
        cell = layer if IS_CUDNN else layer.cell
        return _get_cell_weights(cell, as_tensors, concat_gates)


def _get_cell_weights(rnn_cell, as_tensors=True, concat_gates=True):
    """Retrieves RNN layer weights from their cell(s).
    NOTE: if CuDNNLSTM or CuDNNGRU cell, `rnn_cell` must be the layer instead,
          where non-CuDNN cell attributes are stored.
    """

    def _get_cell_info(rnn_cell):
        rnn_type = type(rnn_cell).__name__.replace('Cell', '')

        if rnn_type in ['SimpleRNN', 'IndRNN']:
            gate_names = ['']
        elif rnn_type in ['LSTM', 'CuDNNLSTM']:
            gate_names = ['i', 'f', 'c', 'o']
        elif rnn_type in ['GRU',  'CuDNNGRU']:
            gate_names = ['z', 'r', 'h']

        if ('CuDNN' in rnn_type) or rnn_cell.use_bias:
            kernel_types = ['kernel', 'recurrent_kernel', 'bias']
        else:
            kernel_types = ['kernel', 'recurrent_kernel']

        return rnn_type, gate_names, kernel_types

    rnn_type, gate_names, kernel_types = _get_cell_info(rnn_cell)

    if TF_KERAS and not concat_gates:
        print(WARN, "getting weights per-gate not supported for tf.keras "
              "implementations; fetching per concat_gates==True instead")
        concat_gates = True
    if not concat_gates and gate_names[0]=='':
        print(WARN, rnn_type + " is not a gated RNN; fetching per "
              "concat_gates==True instead")
        concat_gates = True

    if concat_gates:
        if as_tensors:
            return [getattr(rnn_cell, w_type) for w_type in kernel_types]
        try:
            return rnn_cell.get_weights()
        except:
            return K.batch_get_value(rnn_cell.weights)

    if 'GRU' in rnn_type:
        kernel_types = ['kernel', 'recurrent_kernel', 'input_bias']
    rnn_weights = []
    for w_type in kernel_types:
        rnn_weights.append([])
        for g_name in gate_names:
            rnn_weights[-1].append(getattr(rnn_cell, w_type + '_' + g_name))

    if as_tensors:
        return rnn_weights
    else:
        for weight_idx in range(len(rnn_weights)):
            for gate_idx in range(len(rnn_weights[weight_idx])):
                rnn_weights[weight_idx][gate_idx] = K.eval(
                    rnn_weights[weight_idx][gate_idx])
        return rnn_weights


def rnn_summary(layer):
    """Prints passed RNN layer's weights, and if applicable, gates information
    NOTE: will not print gates information for tf.keras imports as they
    lack pertinent attributes.
    """

    if hasattr(layer, 'backward_layer'):
        rnn_cells = layer.forward_layer, layer.backward_layer
        IS_CUDNN = "CuDNN" in type(rnn_cells[0]).__name__
        if not IS_CUDNN:
            rnn_cells = [layer.cell for layer in rnn_cells]
    else:
        IS_CUDNN = "CuDNN" in type(layer).__name__
        rnn_cells = [layer] if IS_CUDNN else [layer.cell]

    for idx, rnn_cell in enumerate(rnn_cells):
        if len(rnn_cells) == 2:
            if idx == 0:
                print("// FORWARD LAYER")
            elif idx == 1:
                print("// BACKWARD LAYER")

        kernel_types = ['kernel', 'recurrent_kernel', 'bias']
        if type(layer).__name__ == 'GRU':
            kernel_types += ['input_bias']

        for kernel_type in kernel_types:
            weight_matrix = getattr(rnn_cell, kernel_type, None)
            if weight_matrix is not None:
                print(weight_matrix.name, "-- shape=%s" % weight_matrix.shape)

            if not TF_KERAS:
                [print(key, "-- shape=%s" % val.shape) for key, val in
                    rnn_cell.__dict__.items() if (kernel_type + '_' in key)
                    and (len(key) == len(kernel_type) + 2)]
                print()



from . import visuals_gen
from . import visuals_rnn
from . import inspect_gen
from . import inspect_rnn

from .visuals_gen import *
from .visuals_rnn import *
from .inspect_gen import *
from .inspect_rnn import *

__version__ = '1.12'



import numpy as np

from copy import deepcopy
from .utils import _validate_args
from ._backend import K, TF_KERAS


def get_outputs(model, _id, input_data, layer=None, learning_phase=0,
                as_dict=False):
    """Retrieves layer outputs given input data and layer info.

    Arguments:
        model: keras.Model/tf.keras.Model.
        _id: str/int/(list of str/int). int -> idx; str -> name
            idx: int. Index of layer to fetch, via model.layers[idx].
            name: str. Name of layer (full or substring) to be fetched.
                       Returns earliest match if multiple found.
            list of str/int -> treat each str element as name, int as idx.
                      Ex: ['gru', 2] gets outputs of first layer with name
                      substring 'gru', then of layer w/ idx 2
            '*': wildcard -> get outputs of all layers (except input) with
                      'output' attribute. Overrides `layer`.
        input_data: np.ndarray & supported formats(1). Data w.r.t. which loss is
               to be computed for the gradient. Only for mode=='grads'.
        layer: keras.Layer/tf.keras.Layer. Layer whose outputs to return.
               Overrides `_id`.
        learning_phase: bool. 1: use model in train mode
                              0: use model in inference mode
        as_dict: bool. True:  return output fullname-value pairs in a dict
                       False: return output values as list in order fetched

    Returns:
        Layer output values or name-value pairs (see `as_dict`).

    (1): tf.data.Dataset, generators, .tfrecords, & other supported TensorFlow
         input data formats
    """
    def _get_outs_tensors(model, names, idxs, layers):
        if layers is None:
            _id = [x for var in (names, idxs) if var for x in var] or None
            layers = get_layer(model, _id)
        if not isinstance(layers, list):
            layers = [layers]
        return [l.output for l in layers]

    if _id != '*':
        names, idxs, layers, one_requested = _validate_args(_id, layer)
    else:
        # exclude input layer & non-output layers
        names = [l.name for l in model.layers[1:]
                 if getattr(l, 'output', None) is not None]
        idxs, layers = None, None
        one_requested = len(_id) == 1

    layer_outs = _get_outs_tensors(model, names, idxs, layers)

    if TF_KERAS:
        outs_fn = K.function([model.input], layer_outs)
    else:
        outs_fn = K.function([model.input, K.learning_phase()], layer_outs)

    outs = outs_fn([input_data, learning_phase])
    if as_dict:
        return {get_full_name(model, i): x for i, x in zip(names or idxs, outs)}
    return outs[0] if (one_requested and len(outs) == 1) else outs


def get_gradients(model, _id, input_data, labels, layer=None, mode='outputs',
                  sample_weights=None, learning_phase=0, as_dict=False):
    """Retrieves layer gradients w.r.t. outputs or weights.
    NOTE: gradients will be clipped if `clipvalue` or `clipnorm` were set.

    Arguments:
        model: keras.Model/tf.keras.Model.
        _id: str/int/(list of str/int). int -> idx; str -> name
            idx: int. Index of layer to fetch, via model.layers[idx].
            name: str. Name of layer (full or substring) to be fetched.
                       Returns earliest match if multiple found.
            list of str/int -> treat each str element as name, int as idx.
                      Ex: ['gru', 2] gets gradients of first layer with name
                      substring 'gru', then of layer w/ idx 2
            '*': wildcard -> get outputs of all layers (except input) with:
                      - 'output'  attribute (mode == 'outputs')
                      - 'weights' attribute (mode == 'weights')
                      Overrides `layer`.
        input_data: np.ndarray & supported formats(1). Data w.r.t. which loss is
               to be computed for the gradient.
        labels: np.ndarray & supported formats. Labels w.r.t. which loss is
               to be computed for the gradient.
        layer: keras.Layer/tf.keras.Layer. Layer whose gradients to return.
               Overrides `idx` and `name`.
        mode: str. One of: 'outputs', 'weights'. If former, returns grads
               w.r.t. layer outputs(2) - else, w.r.t. layer trainable weights.
        sample_weights: np.ndarray & supported formats. `sample_weight` kwarg
               to model.fit(), etc., weighting individual sample losses.
        learning_phase: bool. 1: use model in train mode
                              0: use model in inference mode
        as_dict: bool. True:  return gradient fullname-value pairs in a dict
                       False: return gradient values as list in order fetched

    Returns:
        Layer weight gradients or gradient-value pairs (see `as_dict`).

    (1): tf.data.Dataset, generators, .tfrecords, & other supported TensorFlow
         input data formats
    (2): "Outputs" are not necessarily activations. If an Activation layer is
         used, returns the gradients of the target layer's PRE-activations
         instead. To then get grads w.r.t. activations, specify the
         Activation layer.
    """

    def _validate_args_(_id, layer, mode):
        if mode not in ['outputs', 'weights']:
            raise Exception("`mode` must be one of: 'outputs', 'weights'")
        return _validate_args(_id, layer)

    if _id != '*':
        names, idxs, layers, one_requested = _validate_args_(_id, layer, mode)
        _id = [x for var in (names, idxs) if var for x in var] or None
    else:
        # exclude input layer & non-output/weightless layers (`mode`-dependent)
        attr = 'output' if mode == 'outputs' else 'weights'
        _id = [l.name for l in model.layers[1:]
               if getattr(l, attr, None) is not None]
        names = _id
        idxs, layers = None, None
        one_requested = len(_id) == 1

    if layers is None:
        layers = get_layer(model, _id)

    grads_fn = _make_grads_fn(model, layers, mode)
    if TF_KERAS:
        grads = grads_fn([input_data, labels])
    else:
        sample_weights = sample_weights or np.ones(len(input_data))
        grads = grads_fn([input_data, sample_weights, labels, 1])

    if as_dict:
        return {get_full_name(model, i): x for i, x in zip(names or idxs, grads)}
    return grads[0] if (one_requested and len(grads) == 1) else grads


def _make_grads_fn(model, layers, mode='outputs'):
    """Returns gradient computation function w.r.t. layer outputs or weights.
    NOTE: gradients will be clipped if `clipnorm` or `clipvalue` were set.
    """
    def _get_params(layers, mode):
        if not isinstance(layers, list):
            layers = [layers]
        if mode == 'outputs':
            return [l.output for l in layers]
        params = []
        _ = [params.extend(l.trainable_weights) for l in layers]
        return params

    if mode not in ('outputs', 'weights'):
        raise Exception("`mode` must be one of: 'outputs', 'weights'")

    params = _get_params(layers, mode)
    grads = model.optimizer.get_gradients(model.total_loss, params)

    if TF_KERAS:
        inputs = [model.inputs[0], model._feed_targets[0]]
    else:
        inputs = [model.inputs[0], model.sample_weights[0],
                  model._feed_targets[0], K.learning_phase()]
    return K.function(inputs=inputs, outputs=grads)


def get_layer(model, _id):
    """Returns layer by index or name.
    If multiple matches are found, returns earliest.
    """
    names, idxs, _, one_requested = _validate_args(_id)

    layers = []
    if idxs is not None:
        layers = [model.layers[i] for i in idxs]
        if names is None:
            return layers if len(layers) > 1 else layers[0]

    for layer in model.layers:
        for n in names:
            if (n in layer.name):
                layers.append(layer)
                _ = names.pop(names.index(n))  # get at most one per match
                break
            # note that above doesn't avoid duplicates, since `names` doesn't

    if len(layers) == 0:
        raise Exception("no layers found w/ names matching substring(s):",
                        ', '.join(names))
    return layers[0] if one_requested else layers


def get_full_name(model, _id):
    """Given full or partial (substring) layer name, or layer index, or list
    containing either, return complete layer name(s).

    Arguments:
        model: keras.Model / tf.keras.Model.
        _id: str/int/(list of str/int). int -> idx; str -> name
            idx: int. Index of layer to fetch, via model.layers[idx].
            name: str. Name of layer (full or substring) to be fetched.
                       Returns earliest match if multiple found.
            list of str/int -> treat each str element as name, int as idx.
                      Ex: ['gru', 2] gets full names of first layer w/ name
                      substring 'gru', and of layer w/ idx 2.

    Returns:
        Full name of layer specified by `_id`.
    """
    names, idxs, _, one_requested = _validate_args(_id)

    fullnames = []
    if idxs is not None:
        fullnames = [model.layers[i].name for i in idxs]
        if names is None:
            return fullnames[0] if one_requested else fullnames

    for layer in model.layers:
        for n in names:
            if n in layer.name:
                fullnames.append(layer.name)
                _ = names.pop(names.index(n))  # get at most one per match
                break
            # note that above doesn't avoid duplicates, since `names` doesn't

    if len(fullnames) == 0:
        raise Exception(f"layer w/ identifier '{_id}' not found")
    return fullnames[0] if one_requested else fullnames


def get_weights(model, _id, omit_names=None, as_dict=False):
    """Given full or partial (substring) weight name(s), return weight values
    (and corresponding names if as_list=False).

    Arguments:
        model: keras.Model / tf.keras.Model
        _id: str/int/tuple of int/(list of str/int/tuple of int).
                      int/tuple of int -> idx; str -> name
            idx: int/tuple of int. Index of layer weights to fetch.
                       int -> all weights of model.layer[idx]
                       tuple of int -> e.g. (idx, wi0, wi1) -> weights indexed
                       wi0, wi1, of model.layer[idx].
            name: str. Name of layer (full or substring) to be fetched.
                       Returns earliest match if multiple found.
                       Can specify a weight (full or substring) in format
                       {name/weight_name}.
            list of str/int/tuple of int -> treat each str element as name,
                       int/tuple of int as idx. Ex: ['gru', 2, (3, 1, 2)] gets
                       weights of first layer with name substring 'gru', then all
                       weights of layer w/ idx 2, then weights w/ idxs 1 and 2 of
                       layer w/ idx 3.
            '*': wildcard -> get weights of all layers with 'weights' attribute.
        omit_names: str/str list. List of names (can be substring) of weights
                                  to omit from fetching.
        as_dict: bool. True:  return weight fullname-value pairs in a dict
                       False: return weight values as list in order fetched

    Returns:
        Layer weight values or name-value pairs (see `as_dict`).
    """
    def _get_weights_tensors(model, _id):
        def _get_by_idx(model, idx):
            if isinstance(idx, tuple) and len(idx) == 2:
                layer_idx, weight_idxs = idx
            else:
                layer_idx, weight_idxs = idx, None

            layer = model.get_layer(index=layer_idx)
            if weight_idxs is None:
                weight_idxs = list(range(len(layer.weights)))  # get all
            if not isinstance(weight_idxs, (tuple, list)):
                weight_idxs = [weight_idxs]

            return {w.name: w for i, w in enumerate(layer.weights)
                    if i in weight_idxs}

        def _get_by_name(model, name):
            # weight_name == weight part of the full weight name
            if len(name.split('/')) == 2:
                layer_name, weight_name = name.split('/')
            else:
                layer_name, weight_name = name.split('/')[0], None
            layer_name = get_full_name(model, layer_name)
            layer = model.get_layer(name=layer_name)

            if weight_name is not None:
                _weights = {}
                for w in layer.weights:
                    if weight_name in w.name:
                        _weights[w.name] = w
            else:
                _weights = {w.name: w for w in layer.weights}
            if len(_weights) == 0:
                raise Exception(f"weight w/ name '{name}' not found")
            return _weights

        if isinstance(_id, str):
            _weights = _get_by_name(model, _id)
        else:
            _weights = _get_by_idx(model, _id)

        w_names = list(_weights)
        for w_name in w_names:
            if any(to_omit in w_name for to_omit in omit_names):
                del _weights[w_name]
        return _weights

    if _id != '*':
        names, idxs, *_ = _validate_args(_id)
        _ids = [x for var in (names, idxs) if var for x in var] or None
    else:
        # exclude input layer & non-weight layers
        _ids = [l.name for l in model.layers[1:]
                if getattr(l, 'weights', None) not in (None, [])]
    if not isinstance(omit_names, list):
        omit_names = [omit_names] if omit_names else []

    weights = {}
    for _id in _ids:
        weights.update(_get_weights_tensors(model, _id))

    weights = {name: value for name, value in
               zip(weights, K.batch_get_value(list(weights.values())))}
    if as_dict:
        return weights
    weights = list(weights.values())
    return weights[0] if (len(_ids) == 1 and len(_ids) == 1) else weights


def _detect_nans(data):
    data = np.asarray(data).ravel()
    perc_nans = 100 * np.sum(np.isnan(data)) / len(data)
    if perc_nans == 0:
        return None
    if perc_nans < 0.1:
        num_nans = int((perc_nans / 100) * len(data))  # show as quantity
        txt = "{:d}% \nNaNs".format(num_nans)
    else:
        txt = "{:.1f}% \nNaNs".format(perc_nans)  # show as percent
    return txt


def weights_norm(model, _id, _dict=None, stat_fns=(np.max, np.mean),
                 norm_fn=np.square, omit_names=None, axis=-1, verbose=0):
    """Retrieves model layer weight matrix norms, as specified by `norm_fn`.

    Arguments:
        model: keras.Model/tf.keras.Model.
        _id: str/int/tuple of int/(list of str/int/tuple of int).
                      int/tuple of int -> idx; str -> name
            idx: int/tuple of int. Index of layer weights to fetch.
                       int -> all weights of model.layer[idx]
                       tuple of int -> e.g. (idx, wi0, wi1) -> weights indexed
                       wi0, wi1, of model.layer[idx].
            name: str. Name of layer (full or substring) to be fetched.
                       Returns earliest match if multiple found.
                       Can specify a weight (full or substring) in format
                       {name/weight_name}.
            list of str/int/tuple of int -> treat each str element as name,
                       int/tuple of int as idx.  Ex: ['gru', 2, (3, 1, 2)] gets
                       weights of first layer with name substring 'gru', then all
                       weights of layer w/ idx 2, then weights w/ idxs 1 and 2 of
                       layer w/ idx 3.
        _dict: dict/None. If None, returns new dict. If dict, appends to it.
        stat_fns: functions list/tuple. Aggregate statistic to compute from
               normed weights.
        norm_fn: function. Norm transform to apply to weights. Ex:
              - np.square (l2 norm)
              - np.abs    (l1 norm)
        omit_names: str/str list. List of names (can be substring) of weights
                                  to omit from fetching.
        axis: int. Axis w.r.t. which compute the norm (collapsing all others).
        verbose: int/bool, 0/1. 1/True: print norm stats, enumerated by layer
               indices. If `_dict` is not None, print last computed results.

    Returns:
        stats_all: dict. dict of lists containing layer weight norm statistics,
               structured: stats_all[layer_fullname][weight_index][stat_index].

    Applied example: https://stackoverflow.com/q/61481921/10133797
    """
    def _process_args(model, _id, _dict):
        def _get_names(model, _ids):
            _ids_normalized = []
            for _id in _ids:
                _ids_normalized.append(_id[0] if isinstance(_id, tuple) else _id)
            names = get_full_name(model, _ids_normalized)
            if not isinstance(names, list):
                names = [names]
            return names

        _ids = _id if isinstance(_id, list) else [_id]
        names = _get_names(model, _ids)

        if _dict:
            stats_all = deepcopy(_dict)  # do not mutate original dict
        else:
            stats_all = {name: [[]] for name in names}
        return _ids, stats_all, names

    def _print_stats(stats_all, l_idx, l_name):
        def _unpack_layer_stats(stats_all, l_name):
            l_stats = stats_all[l_name]
            stats_flat = []
            for w_stats in l_stats:
                if isinstance(w_stats, list):
                    _ = [stats_flat.append(stat[-1]) for stat in w_stats]
                else:
                    stats_flat.append(w_stats)
            return stats_flat

        txt = "{} "
        for w_stats in stats_all[l_name]:
            txt += "{:.4f}, " * len(w_stats)
            txt = txt.rstrip(", ") + " -- "
        txt = txt.rstrip(" -- ")

        stats_flat = _unpack_layer_stats(stats_all, l_name)
        print(txt.format(l_idx, *stats_flat))

    def _get_layer_norm(stats_all, _id, model, l_name, omit_names, norm_fn,
                        stat_fns, axis):
        def _compute_norm(w, norm_fn, axis=-1):
            axis = axis if axis != -1 else len(w.shape) - 1
            reduction_axes = tuple([ax for ax in range(len(w.shape))
                                    if ax != axis])
            return np.sqrt(np.sum(norm_fn(w), axis=reduction_axes))

        def _append(stats_all, l2_stats, w_idx, l_name):
            if len(stats_all[l_name]) < w_idx + 1:
                stats_all[l_name].append([])
            while len(stats_all[l_name][w_idx]) < len(l2_stats):
                stats_all[l_name][w_idx].append([])

            for stat_idx, stat in enumerate(l2_stats):
                stats_all[l_name][w_idx][stat_idx].append(stat)

        weights = get_weights(model, _id, omit_names)

        for w_idx, w in enumerate(weights):
            l2 = _compute_norm(w, norm_fn, axis)
            l2_stats = [fn(l2) for fn in stat_fns]
            _append(stats_all, l2_stats, w_idx, l_name)
        return stats_all

    _ids, stats_all, names = _process_args(model, _id, _dict)
    for l_idx, layer in enumerate(model.layers):
        if layer.name in names:
            _id = _ids[names.index(layer.name)]
            stats_all = _get_layer_norm(stats_all, _id, model, layer.name,
                                        omit_names, norm_fn, stat_fns, axis)
            if verbose:
                _print_stats(stats_all, l_idx, layer.name)
    return stats_all



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

from ._backend import NOTE


def features_0D(data, marker='o', cmap='bwr', color=None, configs=None, **kwargs):
    """Plots 0D aligned scatterplots in a standalone graph.

    iter == list/tuple (both work)

    Arguments:
        data: np.ndarray, 2D: (samples, channels).
        marker: str. Pyplot kwarg specifying scatter plot marker shape.
        cmap: str. Pyplot cmap (colormap) kwarg for the heatmap. Overridden
              by `color`!=None.
        color: (float iter) iter / str / str iter / None. Pyplot kwarg,
              specifying marker colors in order of drawing. If str/ float iter,
              draws all curves in one color. Overrides `cmap`. If None,
              automatically colors along equally spaced `cmap` gradient intervals.
              Ex: ['red', 'blue']; [[0., .8, 1.], [.2, .5, 0.]] (RGB)
        configs: dict. kwargs to customize various plot schemes:
            'plot':  passed to plt.scatter()
            'title': passed to plt.suptitle()
            'save':  passed to fig.savefig() if `savepath` is not None.

    kwargs:
        w: float. Scale width  of resulting plot by a factor.
        h: float. Scale height of resulting plot by a factor.
        show_borders: bool.  If True, shows boxes around plot(s).
        title_mode:   bool/str. If True, shows generic supertitle.
              If str in ('grads', 'outputs'), shows supertitle tailored to
              `data` dim (2D/3D). If other str, shows `title_mode` as supertitle.
              If False, no title is shown.
        show_y_zero: bool. If True, draws y=0.
        ylims: str ('auto'); float list/tuple. Plot y-limits; if 'auto',
               sets both lims to max of abs(`data`) (such that y=0 is centered).
        savepath: str/None. Path to save resulting figure to. Also see `configs`.
               If None, doesn't save.

    Returns:
        (figs, axes) of generated plots.
    """

    w, h         = kwargs.get('w', 1), kwargs.get('h', 1)
    show_borders = kwargs.get('show_borders', False)
    title_mode   = kwargs.get('title_mode', 'outputs')
    show_y_zero  = kwargs.get('show_y_zero', True)
    ylims        = kwargs.get('ylims', 'auto')
    savepath     = kwargs.get('savepath', None)

    def _process_configs(configs, w, h):
        defaults = {
            'plot':  dict(s=15, linewidth=2),
            'title': dict(weight='bold', fontsize=14),
            'save':  dict(),
            }
        configs = configs or {}
        # override defaults, but keep those not in `configs`
        for key in defaults:
            defaults[key].update(configs.get(key, {}))
        kw = defaults.copy()
        return kw

    def _catch_unknown_kwargs(kwargs):
        allowed_kwargs = ('w', 'h', 'show_borders', 'title_mode', 'show_y_zero',
                          'ylims', 'savepath')
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise Exception("unknown kwarg `%s`" % kwarg)

    def _get_title(data, title_mode):
        feature = "Context-feature"
        context = "Context-units"

        if title_mode in ['grads', 'outputs']:
            feature = "Gradients" if title_mode=='grads' else "Outputs"
            context = "Timesteps"
        return "(%s vs. %s) vs. Channels" % (feature, context)

    _catch_unknown_kwargs(kwargs)
    kw = _process_configs(configs, w, h)
    if data.ndim != 2:
        raise Exception("`data` must be 2D")

    if color is None:
        cmap = cm.get_cmap(cmap)
        cmap_grad = np.linspace(0, 256, len(data[0])).astype('int32')
        color = cmap(cmap_grad)
        color = np.vstack([color] * data.shape[0])
    x = np.ones(data.shape) * np.expand_dims(np.arange(1, len(data) + 1), -1)

    if show_y_zero:
        plt.axhline(0, color='k', linewidth=1)
    plt.scatter(x.ravel(), data.ravel(), marker=marker, color=color,
                **kw['plot'])

    plt.gca().set_xticks(np.arange(1, len(data) + 1), minor=True)
    plt.gca().tick_params(which='minor', length=4)
    if ylims == 'auto':
        ymax = np.max(np.abs(data))
        ymin = -ymax
    else:
        ymin, ymax = ylims
    plt.gca().set_ylim(ymin, ymax)

    if title_mode:
        title = _get_title(data, title_mode)
        plt.title(title, **kw['title'])
    if not show_borders:
        plt.box(None)

    fig, axes = plt.gcf(), plt.gca()
    fig.set_size_inches(12 * w, 4 * h)
    plt.show()
    if savepath is not None:
        fig.savefig(savepath, **kw['save'])
    return fig, axes


def features_1D(data, n_rows=None, annotations='auto', equate_axes=True,
                max_timesteps=None, subplot_samples=False, configs=None,
                **kwargs):
    """Plots 1D curves in a standalone graph or subplot grid.

    Arguments:
        data: np.ndarray, 2D/3D. Data to plot.
              2D -> standalone graph; 3D -> subplot grid.
              3D: (samples, timesteps, channels)
              2D: (timesteps, channels) or (timesteps, samples)
        n_rows: int/None. Number of rows in subplot grid. If None,
              determines automatically, closest to n_rows == n_cols.
        annotations: str list/'auto'/None.
            'auto': annotate each subplot with its index.
             list of str: annotate by indexing into the list.
                          If len(list) < len(data), won't annotate remainder.
             None: don't annotate.
        equate_axes:    bool. If True, subplots will share x- and y-axes limits.
        max_timesteps:  int/None. Max number of timesteps to show per plot.
              If None, keeps original.
        subplot_samples: bool. If True, generates a subplot per dim 0 of `data`
              instead of dim 2 - i.e. (Channels vs. Timesteps) vs. Samples.
        configs: dict. kwargs to customize various plot schemes:
            'plot':    passed to ax.plot();      ax  = subplots axis
            'subplot': passed to fig.subplots(); fig = subplots figure
            'title':   passed to fig.suptitle()
            'tight':   passed to fig.subplots_adjust()
            'annot':   passed to ax.annotate()
            'save':    passed to fig.savefig() if `savepath` is not None.

    iter == list/tuple (both work)
    kwargs:
        w: float. Scale width  of resulting plot by a factor.
        h: float. Scale height of resulting plot by a factor.
        show_borders:  bool.  If True, shows boxes around plot(s).
              Ex: [1, 1] -> show both x- and y-ticks (and their labels).
                  [0, 0] -> hide both.
        show_xy_ticks: int/bool iter. Slot 0 -> x, Slot 1 -> y.
        title_mode: bool/str. If True, shows generic supertitle.
              If str in ('grads', 'outputs'), shows supertitle tailored to
              `data` dim (2D/3D). If other str, shows `title_mode` as supertitle.
              If False, no title is shown.
        show_y_zero: bool. If True, draw y=0 for each plot.
        tight: bool. If True, plots compactly by removing subplot padding.
        borderwidth: float / None. Width of subplot borders.
        color: (float iter) iter / str / str iter / None. Pyplot kwarg,
              specifying curve colors in order of drawing. If str/ float iter,
              draws all curves in one color. If None, default coloring is used.
              Ex: ['red', 'blue']; [[0., .8, 1.], [.2, .5, 0.]] (RGB)
        savepath: str/None. Path to save resulting figure to. Also see `configs`.
               If None, doesn't save.

    Returns:
        (figs, axes) of generated plots.
    """

    w, h          = kwargs.get('w', 1), kwargs.get('h', 1)
    show_borders  = kwargs.get('show_borders', True)
    show_xy_ticks = kwargs.get('show_xy_ticks', [1, 1])
    title_mode    = kwargs.get('title_mode', 'outputs')
    show_y_zero   = kwargs.get('show_y_zero', False)
    tight         = kwargs.get('tight', False)
    borderwidth   = kwargs.get('borderwidth', None)
    color         = kwargs.get('color', None)
    savepath      = kwargs.get('savepath', None)

    def _process_configs(configs, w, h, tight, equate_axes):
        defaults = {
            'plot':    dict(),
            'subplot': dict(sharex=True, sharey=True, dpi=76, figsize=(10, 10)),
            'title':   dict(weight='bold', fontsize=14, y=.93 + .12 * tight),
            'tight':   dict(left=0, right=1, bottom=0, top=1, wspace=0, hspace=0),
            'annot':   dict(weight='bold', fontsize=16, color='g',
                            xy=(.03, .9), xycoords='axes fraction'),
            'save':    dict(),
            }
        configs = configs or {}
        # override defaults, but keep those not in `configs`
        for key in defaults:
            defaults[key].update(configs.get(key, {}))
        kw = defaults.copy()

        if not equate_axes:
            kw['subplot'].update({'sharex': False, 'sharey': False})
        size = kw['subplot']['figsize']
        kw['subplot']['figsize'] = (size[0] * w, size[1] * h)
        return kw

    def _catch_unknown_kwargs(kwargs):
        allowed_kwargs = ('w', 'h', 'show_borders', 'show_xy_ticks',
                          'title_mode', 'show_y_zero', 'tight',
                          'borderwidth', 'color', 'savepath')
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise Exception("unknown kwarg `%s`" % kwarg)

    def _get_title(data, title_mode, subplot_samples):
        feature = "Context-feature"
        context = "Context-units"
        if title_mode in ['grads', 'outputs']:
            feature = "Gradients" if title_mode == 'grads' else "Outputs"
            context = "Timesteps"

        subplot_mode_3d = "vs. Samples) vs. Channels"
        subplot_mode_2d = "vs. Channels"
        if subplot_samples:
            subplot_mode_3d = "vs. Channels) vs. Samples"
            subplot_mode_2d = "vs. Samples"

        if data.ndim == 3 and data.shape[-1] != 1:
            return "((%s vs. %s) %s" % (feature, context, subplot_mode_3d)
        else:
            return "(%s vs. %s) %s" % (feature, context, subplot_mode_2d)

    def _get_feature_outputs(data, ax_idx):
        # ax_idx -> samples if subplot_samples==True, else -> channels
        feature_outputs = []
        for sample in data:
            feature_outputs.append(sample[:max_timesteps, ax_idx - 1])
        return feature_outputs

    def _get_style_info(data, n_rows, color, annotations):
        n_subplots = data.shape[-1]
        n_rows, n_cols = _get_nrows_and_ncols(n_rows, n_subplots)

        if color is None:
            n_colors = len(data) if data.ndim == 3 else 1
            color = [None] * n_colors
        if annotations == 'auto':
            annotations = list(map(str, range(n_subplots)))
        elif annotations is not None:
            # ensure external list is unaffected
            annotations = annotations.copy()
        return n_rows, n_cols, color, annotations

    def _process_data(data):
        if data.ndim not in (2, 3):
            raise Exception("`data` must be 2D or 3D")
        if data.ndim == 2 and subplot_samples:
            print(NOTE, "`subplot_samples` w/ 2D `data` will only change "
                  "title shown, and assumes `data` dims (timesteps, samples)")
        elif subplot_samples:
            data = data.T  # -> (channels, timesteps, samples)
        if data.ndim == 2:
            data = data.T  # -> (samples, timesteps) | (channels, timesteps)
            data = np.expand_dims(data, -1)
        return data

    def _style_axis(ax, kw, show_borders, show_xy_ticks, annotations, xmax):
        ax.axis(xmin=0, xmax=xmax)
        if not show_xy_ticks[0]:
            ax.set_xticks([])
        if not show_xy_ticks[1]:
            ax.set_yticks([])
        if annotations:
            ax.annotate(annotations.pop(0), **kw['annot'])
        if not show_borders:
            ax.set_frame_on(False)

    kw = _process_configs(configs, w, h, tight, equate_axes)
    _catch_unknown_kwargs(kwargs)
    data = _process_data(data)
    n_rows, n_cols, color, annotations = _get_style_info(data, n_rows, color,
                                                         annotations)
    fig, axes = plt.subplots(n_rows, n_cols, **kw['subplot'])
    axes = np.asarray(axes)

    if title_mode:
        title = _get_title(data, title_mode, subplot_samples)
        fig.suptitle(title, **kw['title'])

    for ax_idx, ax in enumerate(axes.flat):
        if show_y_zero:
            ax.axhline(0, color='red')

        feature_outputs = _get_feature_outputs(data, ax_idx)
        for idx, out in enumerate(feature_outputs):
            ax.plot(out, color=color[idx], **kw['plot'])

        _style_axis(ax, kw, show_borders, show_xy_ticks, annotations,
                    xmax=len(feature_outputs[0]))

    if tight:
        fig.subplots_adjust(**kw['tight'])
    if borderwidth is not None:
        for ax in axes.flat:
            [s.set_linewidth(borderwidth) for s in ax.spines.values()]

    plt.show()
    if savepath is not None:
        fig.savefig(savepath, **kw['save'])
    return fig, axes


def features_2D(data, n_rows=None, norm=None, cmap='bwr', reflect_half=False,
                timesteps_xaxis=True, max_timesteps=None, configs=None, **kwargs):
    """Plots 2D heatmaps in a standalone graph or subplot grid.

    iter == list/tuple (both work)

    Arguments:
        data: np.ndarray, 2D/3D. Data to plot.
              2D -> standalone graph; 3D -> subplot grid.
              3D: (samples, timesteps, channels)
              2D: (timesteps, channels)
        n_rows: int/None. Number of rows in subplot grid. If None,
              determines automatically, closest to n_rows == n_cols.
        norm: float iter/'auto'/None. Normalizes colors to range between
              norm==(vmin, vmax), according to `cmap`. If 'auto', sets
              norm = (-v, v), where v = max(abs(data)).
                  Ex: `cmap`='bwr' ('blue white red') -> all
                  values <=vmin and >=vmax will be shown as most intense blue
                  and red, and those exactly in-between are shown as white.
        cmap: str. Pyplot cmap (colormap) kwarg for the heatmap.
        reflect_half: bool. If True, second half of channels dim will be
              flipped about the timesteps dim.
        timesteps_xaxis: bool. If True, the timesteps dim (`data` dim 1)
              if plotted along the x-axis.
        max_timesteps:  int/None. Max number of timesteps to show per plot.
              If None, keeps original.
        configs: dict. kwargs to customize various plot schemes:
            'plot':     passed to fig.hist(); fig = subplots figure
            'subplot':  passed to plt.subplots()
            'title':    passed to fig.suptitle()
            'tight':    passed to fig.subplots_adjust()
            'colorbar': passed to fig.colorbar()
            'save':     passed to fig.savefig() if `savepath` is not None.

    kwargs:
        w: float. Scale width  of resulting plot by a factor.
        h: float. Scale height of resulting plot by a factor.
        show_borders:  bool.  If True, shows boxes around plot(s).
        show_xy_ticks: int/bool iter. Slot 0 -> x, Slot 1 -> y.
              Ex: [1, 1] -> show both x- and y-ticks (and their labels).
                  [0, 0] -> hide both.
        show_colorbar: bool. If True, shows one colorbar next to plot(s).
        title_mode:    bool/str. If True, shows generic supertitle.
              If str in ('grads', 'outputs'), shows supertitle tailored to
              `data` dim (2D/3D). If other str, shows `title_mode` as supertitle.
              If False, no title is shown.
        tight: bool. If True, plots compactly by removing subplot padding.
        channel_axis: int, 0 or -1. `data` axis holding channels/features.
              -1 --> (samples,  timesteps, channels)
              0  --> (channels, timesteps, samples)
        borderwidth: float / None. Width of subplot borders.
        savepath: str/None. Path to save resulting figure to. Also see `configs`.
               If None, doesn't save.

    Returns:
        (figs, axes) of generated plots.
    """

    w, h          = kwargs.get('w', 1), kwargs.get('h', 1)
    show_borders  = kwargs.get('show_borders', True)
    show_xy_ticks = kwargs.get('show_xy_ticks', [1, 1])
    show_colorbar = kwargs.get('show_colorbar', False)
    title_mode    = kwargs.get('title_mode', 'outputs')
    tight         = kwargs.get('tight', False)
    channel_axis  = kwargs.get('channel_axis', -1)
    borderwidth   = kwargs.get('borderwidth', None)
    savepath      = kwargs.get('savepath', None)

    def _process_configs(configs, w, h, tight):
        defaults = {
            'plot':    dict(),
            'subplot': dict(sharex=True, sharey=True, dpi=76, figsize=(10, 10)),
            'title':   dict(weight='bold', fontsize=14, y=.93 + .12 * tight),
            'tight':   dict(left=0, right=1, bottom=0, top=1, wspace=0, hspace=0),
            'colorbar': dict(),
            'save':     dict(),
            }
        configs = configs or {}
        # override defaults, but keep those not in `configs`
        for key in defaults:
            defaults[key].update(configs.get(key, {}))
        kw = defaults.copy()

        size = kw['subplot']['figsize']
        kw['subplot']['figsize'] = (size[0] * w, size[1] * h)
        return kw

    def _catch_unknown_kwargs(kwargs):
        allowed_kwargs = ('w', 'h', 'show_borders', 'show_xy_ticks',
                          'show_colorbar', 'title_mode', 'tight',
                          'channel_axis', 'borderwidth', 'savepath')
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise Exception("unknown kwarg `%s`" % kwarg)

    def _get_title(data, title_mode, timesteps_xaxis, vmin, vmax):
        feature = "Context-feature"
        context = "Context-units"
        context_order = "(%s vs. Channels)" % context
        extra_dim = ""

        if title_mode in ['grads', 'outputs']:
            feature = "Gradients" if title_mode == 'grads' else "Outputs"
            context = "Timesteps"
        if timesteps_xaxis:
            context_order = "(Channels vs. %s)" % context
        if data.ndim == 3:
            extra_dim = ") vs. Samples"
            context_order = "(" + context_order

        norm_txt = "(%.4G, %.4G)" % (vmin, vmax) if (
            vmin is not None) else "auto"
        return "{} vs. {}{} -- norm={}".format(context_order, feature,
                                               extra_dim, norm_txt)

    def _get_style_info(data, n_rows, norm):
        if norm == 'auto':
            v = np.abs(data).max()
            vmin, vmax = (-v, v)
        else:
            vmin, vmax = norm or (None, None)
        n_subplots = len(data) if data.ndim == 3 else 1
        n_rows, n_cols = _get_nrows_and_ncols(n_rows, n_subplots)

        return n_rows, n_cols, vmin, vmax

    def _process_data(data, max_timesteps, reflect_half,
                      timesteps_xaxis, channel_axis):
        if data.ndim not in (2, 3):
            raise Exception("`data` must be 2D or 3D")

        if max_timesteps is not None:
            data = data[..., :max_timesteps, :]

        if reflect_half:
            data = data.copy()  # prevent passed array from changing
            half_chs = data.shape[-1] // 2
            data[..., half_chs:] = np.flip(data[..., half_chs:], axis=0)

        if timesteps_xaxis:
            if data.ndim != 3:
                data = np.expand_dims(data, 0)
            data = np.transpose(data, (0, 2, 1))
        return data

    def _style_axis(ax, kw, show_borders, show_xy_ticks):
        ax.axis('tight')
        if not show_xy_ticks[0]:
            ax.set_xticks([])
        if not show_xy_ticks[1]:
            ax.set_yticks([])
        if not show_borders:
            ax.set_frame_on(False)

    _catch_unknown_kwargs(kwargs)
    kw = _process_configs(configs, w, h, tight)
    data = _process_data(data, max_timesteps, reflect_half,
                         timesteps_xaxis, channel_axis)
    n_rows, n_cols, vmin, vmax = _get_style_info(data, n_rows, norm)

    fig, axes = plt.subplots(n_rows, n_cols, **kw['subplot'])
    axes = np.asarray(axes)

    if title_mode:
        title = _get_title(data, title_mode, timesteps_xaxis, vmin, vmax)
        fig.suptitle(title, **kw['title'])

    for ax_idx, ax in enumerate(axes.flat):
        img = ax.imshow(data[ax_idx], cmap=cmap, vmin=vmin, vmax=vmax,
                        **kw['plot'])
        _style_axis(ax, kw, show_borders, show_xy_ticks)

    if show_colorbar:
        fig.colorbar(img, ax=axes.ravel().tolist(), **kw['colorbar'])
    if tight:
        fig.subplots_adjust(**kw['tight'])
    if borderwidth is not None:
        for ax in axes.flat:
            [s.set_linewidth(borderwidth) for s in ax.spines.values()]

    plt.show()
    if savepath is not None:
        fig.savefig(savepath, **kw['save'])
    return fig, axes


def _get_nrows_and_ncols(n_rows, n_subplots):
    if n_rows is None:
        n_rows = int(np.sqrt(n_subplots))
    n_cols = max(int(n_subplots / n_rows), 1)  # ensure n_cols != 0
    n_rows = int(n_subplots / n_cols)

    while not ((n_subplots / n_cols).is_integer() and
               (n_subplots / n_rows).is_integer()):
        n_cols -= 1
        n_rows = int(n_subplots / n_cols)
    return n_rows, n_cols


def features_hist(data, n_rows='vertical', bins=100, xlims=None, tight=True,
                  annotations='auto', configs=None, **kwargs):
    """Plots histograms in a subplot grid.

    Arguments:
        data: np.ndarray, n-dim. Data to plot.
              (samples, ..., channels), or (samples,), or (channels,)
        n_rows: int/'vertical'/None. Number of rows in subplot grid. If None,
              determines automatically, closest to n_rows == n_cols.
              If 'vertical', sets n_rows = len(data).
        bins: int. Pyplot kwarg to plt.hist(), # of bins.
        xlims: float tuple. x limits to apply to all subplots.
        tight: bool. If True, plt.subplots_adjust (spacing) according to
              configs['tight'], defaulted to minimize intermediate padding.
        annotations: str list/'auto'/None.
            'auto': annotate each subplot with its index.
             list of str: annotate by indexing into the list.
                          If len(list) < len(data), won't annotate remainder.
             None: don't annotate.
        configs: dict. kwargs to customize various plot schemes:
            'plot':    passed partly* to ax.hist() in hist_clipped();
                         include `peaks_to_clip` to adjust ylims with a
                         number of peaks disregarded. *See help(hist_clipped).
                         ax = subplots axis
            'subplot': passed to plt.subplots()
            'title':   passed to fig.suptitle(); fig = subplots figure
            'tight':   passed to fig.subplots_adjust()
            'annot':   passed to ax.annotate()
            'save':    passed to fig.savefig() if `savepath` is not None.

    kwargs:
        w: float. Scale width  of resulting plot by a factor.
        h: float. Scale height of resulting plot by a factor.
        show_borders:  bool.  If True, shows boxes around plot(s).
              Ex: [1, 1] -> show both x- and y-ticks (and their labels).
                  [0, 0] -> hide both.
        show_xy_ticks: int/bool iter. Slot 0 -> x, Slot 1 -> y.
        title: str/None. If not None, show `title` as plt.suptitle.
        borderwidth: float / None. Width of subplot borders.
        savepath: str/None. Path to save resulting figure to. Also see `configs`.
               If None, doesn't save.

    Returns:
        (figs, axes) of generated plots.
    """
    w, h          = kwargs.get('w', 1), kwargs.get('h', 1)
    show_borders  = kwargs.get('show_borders', True)
    show_xy_ticks = kwargs.get('show_xy_ticks', [1, 1])
    title         = kwargs.get('title', None)
    borderwidth   = kwargs.get('borderwidth', None)
    savepath      = kwargs.get('savepath', None)

    def _process_configs(configs, w, h, tight):
        defaults = {
            'plot':    dict(peaks_to_clip=0),
            'subplot': dict(sharex='col', sharey=True, dpi=76, figsize=(10, 10)),
            'title':   dict(weight='bold', fontsize=14, y=.93 + .12 * tight),
            'tight':   dict(left=0, right=1, bottom=0, top=1, wspace=0, hspace=0),
            'annot':   dict(weight='bold', fontsize=14, xy=(.02, .7),
                            xycoords='axes fraction', color='g'),
            'save': dict(),
            }
        configs = configs or {}
        # override defaults, but keep those not in `configs`
        for key in defaults:
            defaults[key].update(configs.get(key, {}))
        kw = defaults.copy()

        size = kw['subplot']['figsize']
        kw['subplot']['figsize'] = (size[0] * w, size[1] * h)
        return kw

    def _catch_unknown_kwargs(kwargs):
        allowed_kwargs = ('w', 'h', 'show_borders', 'show_xy_ticks', 'title',
                          'borderwidth', 'savepath')
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise Exception("unknown kwarg `%s`" % kwarg)

    def _get_style_info(data, n_rows, annotations):
        n_subplots = len(data)
        if n_rows == 'vertical':
            n_rows, n_cols = n_subplots, 1
        else:
            n_rows, n_cols = _get_nrows_and_ncols(n_rows, n_subplots)

        if annotations == 'auto':
            annotations = list(map(str, range(n_subplots)))
        elif annotations is not None:
            # ensure external list is unaffected
            annotations = annotations.copy()
        return n_rows, n_cols, annotations

    def _style_axis(ax, kw, show_borders, show_xy_ticks, xlims, annotations):
        if not show_xy_ticks[0]:
            ax.set_xticks([])
        if not show_xy_ticks[1]:
            ax.set_yticks([])
        if annotations:
            ax.annotate(annotations.pop(0), **kw['annot'])
        if not show_borders:
            ax.set_frame_on(False)
        if xlims is not None:
            ax.set_xlim(*xlims)

    _catch_unknown_kwargs(kwargs)
    kw = _process_configs(configs, w, h, tight)
    n_rows, n_cols, annotations = _get_style_info(data, n_rows, annotations)

    fig, axes = plt.subplots(n_rows, n_cols, **kw['subplot'])
    if n_cols == 1:
        axes = np.expand_dims(axes, -1)
    elif n_rows == 1:
        axes = np.expand_dims(axes, 0)
    if title is not None:
        fig.suptitle(title, **kw['title'])

    for ax_idx, ax in enumerate(axes.flat):
        hist_clipped(data[ax_idx], ax=ax, bins=bins, **kw['plot'])
        _style_axis(ax, kw, show_borders, show_xy_ticks, xlims, annotations)

    if tight:
        fig.subplots_adjust(**kw['tight'])
    if borderwidth is not None:
        for ax in axes.flat:
            [s.set_linewidth(borderwidth) for s in ax.spines.values()]

    plt.show()
    if savepath is not None:
        fig.savefig(savepath, **kw['save'])
    return fig, axes


def features_hist_v2(data, colnames=None, bins=100, xlims=None, ylim=None,
                     tight=True, side_annot=None, configs=None, **kwargs):
    """Plots histograms in a subplot grid; tailored for multiple histograms
    per gridcell.

    Arguments:
        data: np.ndarray / list of np.ndarray / list of lists. Data to plot.
              (rows, cols, subdata...)
              - Plots dim0 along rows, dim1 along cols, subdata within
              individual subplots (iteratively), flattens remaining dims.
              - If list of (list / np.ndarray), each of (top-level) list's
              entries must contain same number of lists / arrays.
              - `subdata` can have any number of entries of any dims, generating
              dynamic number of histograms per gridcell.
        colnames: str list. Column titles, displayed on top subplot boxes.
        bins: int. Pyplot kwarg to plt.hist(), # of bins.
        xlims: float tuple. x limits to apply to all subplots.
        ylim: float. Top y limit of all subplots.
        tight: bool. If True, plt.subplots_adjust (spacing) according to
              configs['tight'], defaulted to minimize intermediate padding.
        side_annot: str. Text to display to the right side of rightmost subplot
              boxes, enumerated by row number ({side_annot}{row})
        configs: dict. kwargs to customize various plot schemes:
            'plot':       passed partly* to ax.hist() in hist_clipped();
                            include `peaks_to_clip` to adjust ylims with a
                            number of peaks disregarded. *See help(hist_clipped).
                            ax = subplots axis
            'subplot':    passed to plt.subplots()
            'title':      passed to fig.suptitle(); fig = subplots figure
            'tight':      passed to fig.subplots_adjust()
            'colnames':   passed to ax.set_title()
            'side_annot': passed to ax.annotate()
            'save':       passed to fig.savefig() if `savepath` is not None.

    kwargs:
        w: float. Scale width  of resulting plot by a factor.
        h: float. Scale height of resulting plot by a factor.
        show_borders:  bool.  If True, shows boxes around plot(s).
              Ex: [1, 1] -> show both x- and y-ticks (and their labels).
                  [0, 0] -> hide both.
        show_xy_ticks: int/bool iter. Slot 0 -> x, Slot 1 -> y.
        title: str/None. If not None, show `title` as plt.suptitle.
        borderwidth: float / None. Width of subplot borders.
        savepath: str/None. Path to save resulting figure to. Also see `configs`.
               If None, doesn't save.

    Returns:
        (figs, axes) of generated plots.
    """
    w, h          = kwargs.get('w', 1), kwargs.get('h', 1)
    show_borders  = kwargs.get('show_borders', True)
    show_xy_ticks = kwargs.get('show_xy_ticks', [1, 1])
    title         = kwargs.get('title', None)
    borderwidth   = kwargs.get('borderwidth', None)
    savepath      = kwargs.get('savepath', None)

    def _process_configs(configs, w, h):
        defaults = {
            'plot':    dict(peaks_to_clip=0),
            'subplot': dict(sharex='col', sharey=True, dpi=76, figsize=(10, 10)),
            'title':   dict(weight='bold', fontsize=15, y=1.06),
            'tight':   dict(left=0, right=1, bottom=0, top=1,
                            wspace=.05, hspace=.05),
            'colnames':   dict(weight='bold', fontsize=14),
            'side_annot': dict(weight='bold', fontsize=14,
                               xy=(1.02, .5), xycoords='axes fraction'),
            'save': dict(),
            }
        configs = configs or {}
        # override defaults, but keep those not in `configs`
        for key in defaults:
            defaults[key].update(configs.get(key, {}))
        kw = defaults.copy()

        size = kw['subplot']['figsize']
        kw['subplot']['figsize'] = (size[0] * w, size[1] * h)
        return kw

    def _catch_unknown_kwargs(kwargs):
        allowed_kwargs = ('w', 'h', 'show_borders', 'show_xy_ticks', 'title',
                          'borderwidth', 'savepath')
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise Exception("unknown kwarg `%s`" % kwarg)

    def _get_data_info(data):
        n_rows = len(data)
        n_cols = len(data[0])
        if isinstance(data, (list, tuple)):
            assert all(len(x) == n_cols for x in data[1:]), (
                "When `data` is a list of lists or list of arrays, each "
                "of the (top-level) list's entries must have the same "
                "number of elements")
        return n_rows, n_cols

    def _style_axis(ax, kw, show_borders, show_xy_ticks, xlims):
        if row == 0 and colnames is not None:
            ax.set_title(f"{colnames[col]}", **kw['colnames'])
        if side_annot is not None and col == n_cols - 1:
            ax.annotate(f"{side_annot}{row}", **kw['side_annot'])
        if not show_borders:
            ax.set_frame_on(False)
        if not show_xy_ticks[0]:
            ax.set_xticks([])
        if not show_xy_ticks[1]:
            ax.set_yticks([])
        if xlims is not None:
            ax.set_xlim(*xlims)

    _catch_unknown_kwargs(kwargs)
    kw = _process_configs(configs, w, h)
    n_rows, n_cols = _get_data_info(data)

    fig, axes = plt.subplots(n_rows, n_cols, **kw['subplot'])
    if n_cols == 1:
        axes = np.expand_dims(axes, -1)
    elif n_rows == 1:
        axes = np.expand_dims(axes, 0)
    if title is not None:
        fig.suptitle(title, **kw['title'])

    for row in range(n_rows):
        for col in range(n_cols):
            ax = axes[row, col]
            for subdata in data[row][col]:
                hist_clipped(subdata, ax=ax, bins=bins, **kw['plot'])

            _style_axis(ax, kw, show_borders, show_xy_ticks, xlims)

    if ylim is not None:
        ax.set_ylim(0, ylim)
    if tight:
        fig.subplots_adjust(**kw['tight'])
    if borderwidth is not None:
        for ax in axes.flat:
            [s.set_linewidth(borderwidth) for s in ax.spines.values()]

    plt.show()
    if savepath is not None:
        fig.savefig(savepath, **kw['save'])
    return fig, axes


def hist_clipped(data, peaks_to_clip=1, ax=None, annot_kw={}, **kw):
    """Histogram with ymax adjusted to a sub-maximal peak, with excluded peaks
    annotated. Useful when extreme peaks dominate plots.

    Arguments:
        data. np.ndarray / list. Data to plot.
        peaks_to_clip: int. ==0 -> regular histogram.
                             >0 -> (e.g. 2) adjust ymax to third peak.
        ax: AxesSubplot. To plot via ax.hist instead of plt.hist.
        annot_kw: dict/None. Passed to plt.annotate().
           None: doesn't annotate.
           {}: uses defaults.
           Non-empty dict: fills with defaults where absent.
        **kw: dict. Passed to plt.hist().

    Returns:
        None
    """
    def _annotate(ax, peaks_info, annot_kw):
        def _process_annot_kw(annot_kw):
            defaults = dict(weight='bold', fontsize=13, color='r',
                            xy=(.83, .85), xycoords='axes fraction')
            if not annot_kw:
                annot_kw = defaults.copy()
            else:
                annot_kw = annot_kw.copy()  # ensure external dict unaffected
                # if `defaults` key not in `annot_kw`, add it & its value
                for k, v in defaults.items():
                    if k not in annot_kw:
                        annot_kw[k] = v
            return annot_kw

        def _make_annotation(peaks_info):
            txt = ''
            for entry in peaks_info:
                txt += "({:.2f}, {})\n".format(entry[0], int(entry[1]))
            return txt.rstrip('\n')

        annot_kw = _process_annot_kw(annot_kw)
        txt = _make_annotation(peaks_info)
        if ax is not None:
            ax.annotate(txt, **annot_kw)
        else:
            plt.annotate(txt, **annot_kw)

    if ax is not None:
        N, bins, _ = ax.hist(np.asarray(data).ravel(), **kw)
    else:
        N, bins, _ = plt.hist(np.asarray(data).ravel(), **kw)

    if peaks_to_clip == 0:
        return

    Ns = np.sort(N)
    lower_max = Ns[-(peaks_to_clip + 1)]

    peaks_info = []
    for peak_idx in range(1, peaks_to_clip + 1):
        patch_idx = np.where(N == Ns[-peak_idx])[0][0]
        peaks_info.append([bins[patch_idx], N[patch_idx]])

    if ax is not None:
        ax.set_ylim(0, lower_max * 1.05)  # include small gap
    else:
        plt.ylim(0, lower_max * 1.05)

    if annot_kw is not None:
        _annotate(ax, peaks_info, annot_kw)



import os
import re
from setuptools import setup, find_packages

current_path = os.path.abspath(os.path.dirname(__file__))


def read_file(*parts):
    with open(os.path.join(current_path, *parts), encoding='utf-8') as reader:
        return reader.read()


def get_requirements(*parts):
    with open(os.path.join(current_path, *parts), encoding='utf-8') as reader:
        return list(map(lambda x: x.strip(), reader.readlines()))


def find_version(*file_paths):
    version_file = read_file(*file_paths)
    version_matched = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                version_file, re.M)
    if version_matched:
        return version_matched.group(1)
    raise RuntimeError('Unable to find version')


setup(
    name="see-rnn",
    version=find_version('see_rnn', '__init__.py'),
    packages=find_packages(exclude=['tests']),
    url="https://github.com/OverLordGoldDragon/see-rnn",
    license="MIT",
    author="OverLordGoldDragon",
    author_email="16495490+OverLordGoldDragon@users.noreply.github.com",
    description=('RNN weights, gradients, & activations visualization '
                 'in Keras & TensorFlow'),
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    keywords=(
        "rnn tensorflow keras visualization deep-learning lstm gru "
    ),
    install_requires=get_requirements('requirements.txt'),
    include_package_data=True,
    zip_safe=True,
    tests_require=["pytest>=4.0", "pytest-cov"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
